@@ -20,8 +20,8 @@
  * @author Sam Lang, Sam Meder
  *
  * $RCSfile: globus_gsi_cred_handle.c,v $
- * $Revision: 1.35.4.2 $
- * $Date: 2008/09/11 12:52:50 $
+ * $Revision: 1.38.4.1 $
+ * $Date: 2010/02/19 21:18:49 $
  */
 #endif
 
@@ -1279,7 +1279,11 @@ globus_gsi_cred_get_policies(
         goto exit;
     }
 
+#if OPENSSL_VERSION_NUMBER < 0x10000000L 
     if((*policies = sk_new_null()) == NULL)
+#else
+    if((*policies = sk_OPENSSL_STRING_new_null()) == NULL)
+#endif
     {
         GLOBUS_GSI_CRED_OPENSSL_ERROR_RESULT(
             result,
@@ -1358,7 +1362,11 @@ globus_gsi_cred_get_policies(
             goto error_exit;
         }
 
+#if OPENSSL_VERSION_NUMBER < 0x10000000L 
         if(sk_push(*policies, final_policy_string) == 0)
+#else
+        if(sk_OPENSSL_STRING_push(*policies, (OPENSSL_STRING)final_policy_string) == 0)
+#endif
         {
             GLOBUS_GSI_CRED_OPENSSL_ERROR_RESULT(
                 result,
@@ -1390,7 +1398,11 @@ globus_gsi_cred_get_policies(
 
     if(*policies != NULL)
     {
+#if OPENSSL_VERSION_NUMBER < 0x10000000L 
         sk_pop_free(*policies, free);
+#else
+        sk_OPENSSL_STRING_pop_free(*policies, free);
+#endif
     }
     *policies = NULL;
     
@@ -1525,6 +1537,71 @@ globus_gsi_cred_get_policy_languages(
 }
 /* @} */
 
+/**
+ * @name Get Cred Cert X509 Issuer Name object
+ */
+/* @{ */
+/**
+ * @ingroup globus_gsi_cred_handle
+ * Get the credential handle's certificate issuer name
+ *
+ * @param handle
+ *        The credential handle containing the certificate
+ *        to get the issuer name of
+ * @param issuer_name
+ *        The issuer name as an X509_NAME object.  This should be freed
+ *        using X509_NAME_free when the user is finished with it
+ * @return 
+ *        GLOBUS_SUCCESS if no error, a error object id otherwise
+ */
+globus_result_t globus_gsi_cred_get_X509_issuer_name(
+    globus_gsi_cred_handle_t            handle,
+    X509_NAME **                        issuer_name)
+{
+    globus_result_t                     result;
+    static char *                       _function_name_ =
+        "globus_gsi_cred_get_issuer_name";
+
+    GLOBUS_I_GSI_CRED_DEBUG_ENTER;
+
+    if(handle == NULL)
+    {
+        GLOBUS_GSI_CRED_ERROR_RESULT(
+            result,
+            GLOBUS_GSI_CRED_ERROR_WITH_CRED,
+            (_GCRSL("NULL cred handle passed to function: %s"), _function_name_));
+        goto error_exit;
+    }
+
+    if(issuer_name == NULL)
+    {
+        GLOBUS_GSI_CRED_ERROR_RESULT(
+            result,
+            GLOBUS_GSI_CRED_ERROR_WITH_CRED,
+            (_GCRSL("NULL issuer name parameter passed to function: %s"), 
+             _function_name_));
+        goto error_exit;
+    }
+
+    if((*issuer_name = 
+        X509_NAME_dup(X509_get_issuer_name(handle->cert))) == NULL)
+    {
+        GLOBUS_GSI_CRED_OPENSSL_ERROR_RESULT(
+            result,
+            GLOBUS_GSI_CRED_ERROR_WITH_CRED_CERT,
+            (_GCRSL("Couldn't get issuer name of credential's cert")));
+        goto error_exit;
+    }
+
+    result = GLOBUS_SUCCESS;
+
+ error_exit:
+
+    GLOBUS_I_GSI_CRED_DEBUG_EXIT;
+    return result;
+}
+/* @} */
+
 /**
  * @name Get Issuer Name
  */
